//-----------------------------------------------
// CMU-800 MIDIKIT AVR Version by RJB
// Device   : ATmega168-20PI
// X'tal    : 20MHz
// FUSEݒ : -FL11101111
// Version  : 0.1alfa Dec. 16 2007
//            0.2beta Dec. 30 2007				
//            0.9beta Jan. 02 2008				
//            1.0     Jan. 11 2008				
//-----------------------------------------------
// include

#include <avr\io.h>
#include <avr\signal.h>
#include <avr\interrupt.h>

//-----------------------------------------------
// define

#define PCB2

#define	A8253CTL1	0x03
#define	A8253CTL2	0x07
#define	A8255PA		0x08
#define	A8255PB		0x09
#define	A8255PC		0x0A
#define	A8255CTL	0x0B

#ifdef PCB2

	#define	WR			PD7
	#define WR_DISABLE		(PORTD |= (1 << WR))
	#define WR_ENABLE		(PORTD &= ~(1 << WR))

	#define SYNC_CLK	PD5
	#define SYNC_START	PD6
	#define SYNC_CLK_HIGH	(PORTD |= (1 << SYNC_CLK))
	#define SYNC_CLK_LOW	(PORTD &= ~(1 << SYNC_CLK))
	#define SYNC_RUN		(PORTD |= (1 << SYNC_START))
	#define SYNC_STOP		(PORTD &= ~(1 << SYNC_START))

#else

	#define	WR			PC4
	#define WR_DISABLE		(PORTC |= (1 << WR))
	#define WR_ENABLE		(PORTC &= ~(1 << WR))

	#define SYNC_CLK	PD4	
	#define SYNC_START	PD5
	#define SYNC_CLK_HIGH	(PORTD |= (1 << SYNC_CLK))
	#define SYNC_CLK_LOW	(PORTD &= ~(1 << SYNC_CLK))
	#define SYNC_RUN		(PORTD |= (1 << SYNC_START))
	#define SYNC_STOP		(PORTD &= ~(1 << SYNC_START))

#endif

#define LED_MIDI	PD2
#define LED_BEAT	PD3
#define LED_MIDI_OFF	(PORTD |= (1 << LED_MIDI))
#define LED_MIDI_ON		(PORTD &= ~(1 << LED_MIDI))
#define LED_BEAT_OFF	(PORTD |= (1 << LED_BEAT))
#define LED_BEAT_ON		(PORTD &= ~(1 << LED_BEAT))

#define	TRUE			1
#define	FALSE			0
#define ON				1
#define OFF				0
#define RHYTHMOFF		0xFF

#define MAX_CH			16
#define MAX_NOTE_CNT	8
#define	INTRCOUNT		0xFEC8	// ^C} 1ms
#define	BUFSIZE			16		// Oobt@TCY

// MIDIbZ[W
#define MIDI_NoteOff			0x80
#define MIDI_NoteOn				0x90
#define MIDI_PolykeyPressure	0xA0
#define MIDI_ControlChange		0xB0
#define MIDI_ProgramChange		0xC0
#define MIDI_ChannelPressure	0xD0
#define MIDI_PitchBend			0xE0

#define MIDI_StartSysEx			0xF0
#define MIDI_TuneRequest		0xF6
#define MIDI_EndSysEx			0xF7

#define MIDI_TimingClock		0xF8
#define MIDI_Start				0xFA
#define MIDI_Continue			0xFB
#define MIDI_Stop				0xFC
#define MIDI_ActiveSensing		0xFE
#define MIDI_SystemReset		0xFF

#define MIDI_CC_Moduration		0x01
#define MIDI_CC_DataEntry		0x06
#define MIDI_CC_Control1		0x10 // MOD SHAPE
#define MIDI_CC_Control2		0x11 // MOD RATE
#define MIDI_CC_Control3		0x12 // MOD DEPTH
#define MIDI_CC_RPN_LSB			0x64
#define MIDI_CC_RPN_MSB			0x65

#define MIDI_MM_AllSoundOff		0x78
#define MIDI_MM_ResetAllControl	0x79
#define MIDI_MM_AllNoteOff		0x7B

//-----------------------------------------------
// f[^^`

typedef char			schar;
typedef unsigned char	uchar;
typedef	int				int16;
typedef	unsigned int	uint16;
typedef	long			int32;
typedef	unsigned long	uint32;
typedef	unsigned char	BOOL;

//-----------------------------------------------
// O[oϐ

uchar	RxByte;
uchar	PC;
uchar	SysEx;
uchar	Rch;
uchar	MByte[2];
uchar	CkCnt;
uchar	CkTim;
uint16	CkPhase;
BOOL	Setflg;

uchar	CVBuf[MAX_CH];
uchar	PBBuf[MAX_CH];
uchar	MWBuf[MAX_CH];
uchar	RateBuf[MAX_CH];
uchar	ShapeBuf[MAX_CH];
uchar	DepthBuf[MAX_CH];
uchar	RPNMsbBuf[MAX_CH];
uchar	RPNLSBBuf[MAX_CH];
uchar	DetuneBuf[MAX_CH];
uchar	NoteCnt[MAX_CH];
uchar	NoteBuf[MAX_CH][MAX_NOTE_CNT];
uchar	PlayNoteBuf[MAX_CH];

// Oobt@p
uchar	RxBuf[BUFSIZE];
int16	Ptr_buf_in, Ptr_buf_out; // |C^

// ݒpe[u
const uint16	DivBase[12] = {38711, 36539, 34488, 32552, 30725, 29001, 27373, 25837, 24386, 23018, 21726, 20506};
const uchar		DetuneBase[12] = {35, 33, 32, 30, 28, 26, 25, 24, 22, 21, 20, 19};
const uchar 	BendBase[13] = {148, 139, 131, 124, 117, 110, 104, 98, 93, 88, 83, 78, 74};
const uchar 	ModBase[12] = {9, 8, 8, 7, 7, 6, 6, 6,	5, 5, 5, 5};

#ifdef PCB2
// 6bit MSB-LSB ]}gbNX
const uchar 	RMtx[64] = { 0x00, 0x20, 0x10, 0x30, 0x08, 0x28, 0x18, 0x38,
							 0x04, 0x24, 0x14, 0x34, 0x0C, 0x2C, 0x1C, 0x3C,
							 0x02, 0x22, 0x12, 0x32, 0x0A, 0x2A, 0x1A, 0x3A,
							 0x06, 0x26, 0x16, 0x36, 0x0E, 0x2E, 0x1E, 0x3E,
							 0x01, 0x21, 0x11, 0x31, 0x09, 0x29, 0x19, 0x39,
							 0x05, 0x25, 0x15, 0x35, 0x0D, 0x2D, 0x1D, 0x3D,
							 0x03, 0x23, 0x13, 0x33, 0x0B, 0x2B, 0x1B, 0x3B,
							 0x07, 0x27, 0x17, 0x37, 0x0F, 0x2F, 0x1F, 0x3F};
#endif

//-----------------------------------------------
// USART荞

SIGNAL(SIG_USART_RECV)
{
	// Oobt@֕ۑ
	Ptr_buf_in++;
	Ptr_buf_in &= (BUFSIZE - 1);
	RxBuf[Ptr_buf_in] = UDR0;
}

//-----------------------------------------------
// ^C}1I[o[t[荞

SIGNAL(SIG_OVERFLOW1)
{
	TCNT1 = INTRCOUNT;

	CkPhase ++;
	Setflg = TRUE;

	if(CkTim++ > 4) SYNC_CLK_LOW;
}

//-----------------------------------------------
// EGCg

void Wait(int32 i)
{
	while(i-- > 0);
}

//-----------------------------------------------
// CMU-800֏o

void OutPort(uchar addr, uchar data)
{
#ifdef PCB2

	PORTB = (PORTB & 0xC0) | RMtx[(((addr << 2) & 0x3C) | (data >> 6))];
	PORTC = (PORTC & 0xC0) | RMtx[(data & 0x3F)];

#else

	PORTC = (PORTC & 0xF0) | ( addr & 0x0F);
	PORTB = (PORTB & 0xC0) | ( data & 0x3F);
	PORTD = (PORTD & 0x3F) | ( data & 0xC0);

#endif

	WR_ENABLE;
	WR_DISABLE;
}

//-----------------------------------------------
// 8255

void Init8255(void)
{
	OutPort(A8255CTL, 0x88);
	OutPort(A8255PA, 0x80);
	OutPort(A8255PB, 0xFF);
	OutPort(A8255PC, 0xFF);
}

//-----------------------------------------------
// 8253

void Init8253(void)
{
	// 8253 #1 CH1-CH3
	OutPort(A8253CTL1, 0x36);
	OutPort(A8253CTL1, 0x76);
	OutPort(A8253CTL1, 0xB6);

	// 8253 #2 CH4-CH6
	OutPort(A8253CTL2, 0x36);
	OutPort(A8253CTL2, 0x76);
	OutPort(A8253CTL2, 0xB6);
}

//-----------------------------------------------
// GATEo

void Gate(uchar onoff)
{
	if(Rch > 0x07)	return;

	OutPort(A8255PC, (Rch << 1) + 1);

	if(onoff) {
		CVBuf[Rch] &= ~0x80;
	}
	else {
		CVBuf[Rch] |= 0x80;
	}
	OutPort(A8255PA, CVBuf[Rch]);

	OutPort(A8255PC, Rch << 1);
	Wait(60);	// Wait 50us
	OutPort(A8255PC, (Rch << 1) + 1);
}

//-----------------------------------------------
// SGATEo͂OFF

void AllGateOff(void)
{
	uchar ch;

	for(ch = 0; ch < 8; ch++) {
		OutPort(A8255PC, (ch << 1) + 1);
		OutPort(A8255PA, 0x80);

		OutPort(A8255PC, ch << 1);
		Wait(60);	// Wait 50us
		OutPort(A8255PC, (ch << 1) + 1);
	}
}

//-----------------------------------------------
// CVo͓dZbg

void SetCV(uchar ch)
{
	uchar note;

	note = PlayNoteBuf[ch];
	if( note < 24) {
		note = 24;
	}
	note -= 24;

	while(note > 63) {
		note -= 12;
	}

	OutPort(A8255PC, (ch << 1) + 1);

	CVBuf[ch] = (CVBuf[ch] & 0x80) | (note & 0x3F);
	OutPort(A8255PA, CVBuf[ch]);

	OutPort(A8255PC, ch << 1);
	Wait(60);	// Wait 50us
	OutPort(A8255PC, (ch << 1) + 1);
}

//-----------------------------------------------
// DCOgZbg

void SetDCO(uchar ch)
{
	uchar	note,ctn,val,ph;
	schar	lfo;
	uint16	m,n,div,bend;
	int16	mod,detune;

	if(ch > 0x05) return;
	note = PlayNoteBuf[ch];

	ctn = ch; 
	if( ctn > 2) ctn++;

	// {
	if(note < 24) note = 24;
	note -= 24;
	m = note / 12;
	n = note % 12;
	div = DivBase[n] >> m;

	// f`[ʂɂ镪ω
	detune = (DetuneBuf[ch] - 64) * DetuneBase[n];

	// W[Vg
	if(RateBuf[ch] < 3) 	ph = CkPhase >> (3 - RateBuf[ch]); 
	else 					ph = CkPhase << (RateBuf[ch] - 3); 

	// W[Vg`
	switch (ShapeBuf[ch]) {
		case 0:	// Opg
			if(ph > 128)	lfo = 191 - ph;	
			else			lfo = ph - 64;
			break;
		case 1: // mRMg
			lfo = 63 - (ph >> 1);
			break;
		case 2: // tmRMg
			lfo = (ph >> 1) - 64;
			break;
		case 3: // `g	
			if(ph > 128)	lfo = -64;	
			else			lfo = 63;
			break;
	}
	// W[Vʂɂ镪ω
	mod = (MWBuf[ch] * ModBase[n] * lfo) >> (4 - DepthBuf[ch]);

	// sb`xhʂŏI
	val = PBBuf[ch];
	if(val < 64) {
		bend = (BendBase[n] * (64 - val)) >> 1;
		div = (DivBase[n] - detune + bend + mod) >> m;
	}
	else {
		bend = (BendBase[n+1] * (val - 64)) >> 1;
		div = (DivBase[n] - detune - bend + mod) >> m;
	}

	// 8253ɃZbg
	OutPort(ctn, (div & 0xFF));	// LSB
	OutPort(ctn, (div >> 8));	// MSB
} 

//-----------------------------------------------
// YgK[

void RhythmTrig(uchar note)
{
	uchar d;

	switch(note){
		case 35: d = 0x80; break; // ABD
		case 36: d = 0x80; break; // BD
		case 40: d = 0x40; break; // ESD
		case 38: d = 0x40; break; // SD
		case 41: d = 0x20; break; // LFT
		case 43: d = 0x20; break; // HFT
		case 45: d = 0x20; break; // LT
		case 47: d = 0x20; break; // LMT
		case 48: d = 0x10; break; // HMT
		case 50: d = 0x10; break; // HT
		case 49: d = 0x08; break; // CR1
		case 57: d = 0x08; break; // CR2
		case 46: d = 0x04; break; // OH
		case 42: d = 0x02; break; // CH
		default: d = 0x00; break;
	}
	OutPort(A8255PB, ~d);
}

//-----------------------------------------------
// m[gON

void NoteON(void)
{
	uchar i;
	uchar max = 0;

	// MIDI9chȏ͖
	if(Rch > 0x07) return;

	// obt@ɓo^ς̃m[g͖
	for(i = 0; i < NoteCnt[Rch]; i++) {
		if(NoteBuf[Rch][i] == MByte[0]) {
			return;
		}
	}

	// obt@ςH
	if(NoteCnt[Rch] == MAX_NOTE_CNT) {
		// ʓ˂
		for(i = 0; i < (MAX_NOTE_CNT - 1); i++) {
			NoteBuf[Rch][i] = NoteBuf[Rch][i+1];
		}
		NoteBuf[Rch][MAX_NOTE_CNT - 1] = MByte[0];
	}
	else {
		NoteBuf[Rch][NoteCnt[Rch]] = MByte[0];
		NoteCnt[Rch]++;
	}

	// ō
	for(i = 0; i < NoteCnt[Rch]; i++) {
		if(max < NoteBuf[Rch][i]) {
			max = NoteBuf[Rch][i];
		}
	}
	PlayNoteBuf[Rch] = max;

	SetCV(Rch);
	SetDCO(Rch);
	Gate(ON);
}

//-----------------------------------------------
// m[gOFF

void NoteOFF(void)
{
	uchar i;
	uchar max = 0;
	BOOL flg = FALSE;

	// MIDI9chȏ͖
	if(Rch > 0x07) return;

	// obt@ɓo^Ăꍇ͍폜
	for(i = 0; i < NoteCnt[Rch]; i++) {
		if(flg) {
			NoteBuf[Rch][i-1] = NoteBuf[Rch][i];
		}
		if(NoteBuf[Rch][i] == MByte[0]) {
			flg = TRUE;
		}
	}
	if(flg) NoteCnt[Rch]--;

	if(NoteCnt[Rch] == 0) {
		// obt@J̏ꍇGATEIt
		Gate(OFF);
	}
	else {
		// ō
		for(i = 0; i < NoteCnt[Rch]; i++) {
			if(max < NoteBuf[Rch][i]) {
				max = NoteBuf[Rch][i];
			}
		}
		PlayNoteBuf[Rch] = max;

		SetCV(Rch);
		SetDCO(Rch);
	}
}

//-----------------------------------------------
// MIDIVXeA^CbZ[W

void MIDI_SystemRealtimeMessage(void)
{
	switch(RxByte) {
		case MIDI_TimingClock:
			SYNC_CLK_HIGH;
			CkTim = 0;
			CkCnt++;
			if(CkCnt > 23) {
				CkCnt = 0;
				LED_BEAT_ON;
			}
			if(CkCnt == 5) {
				LED_BEAT_OFF;
			}
			break;
		case MIDI_Start:
			CkCnt = 0;
			SYNC_RUN;
			break;
		case MIDI_Continue:
			SYNC_RUN;
			break;
		case MIDI_Stop:
			SYNC_STOP;
			break;
		case MIDI_ActiveSensing:
			break;
		case MIDI_SystemReset:
			RhythmTrig(RHYTHMOFF);
			AllGateOff();
			break;
	}
}

//-----------------------------------------------
// MIDIVXebZ[W

BOOL MIDI_SystemMessage(void)
{
	if(SysEx){
		if(RxByte == MIDI_EndSysEx){
			SysEx = FALSE;
		}
	}
	else{
		if(RxByte < 0xF8){
			if(RxByte > 0x7F){
				if(RxByte == MIDI_StartSysEx){
					SysEx = TRUE;
				}
				else{
					Rch = RxByte & 0x0F;
                }
				PC = 0;
 			}
			else{
				MByte[PC & 0x01] = RxByte;
			}
			return TRUE;
		}
		else {
			MIDI_SystemRealtimeMessage();
		}
	}
	return FALSE;
}

//-----------------------------------------------
// MIDI`lbZ[W

void MIDI_ChannelMessage(void)
{
	switch(PC){
		case 0:
			switch(RxByte & 0xF0){
			case MIDI_NoteOff:
				PC = 2;
				break;
			case MIDI_NoteOn:
				PC = 4;
				break;
			case MIDI_PolykeyPressure:
				PC = 6;
				break;
			case MIDI_ProgramChange:
				PC = 8;
				break;
			case MIDI_ControlChange:
				PC = 10;
				break;
			case MIDI_ChannelPressure:
				PC = 12;
				break;
			case MIDI_PitchBend:
				PC = 14;
				break;
			} break;

        // m[gIt
		case 2:
			PC = 3;
			break;
        case 3:
			PC = 2;
			NoteOFF();
			RhythmTrig(RHYTHMOFF);
			LED_MIDI_OFF;
			break;

        // m[gI
        case 4:
			PC = 5;
			break;
        case 5:
			PC = 4;
			if( MByte[1] == 0){
				NoteOFF();
				RhythmTrig(RHYTHMOFF);
				LED_MIDI_OFF;
			}
			else{
				NoteON();
				if(Rch == 0x09) {
					RhythmTrig(MByte[0]);
				}
				LED_MIDI_ON;
			}
			break;

		// |tHjbNL[vbV[ 
        case 6:
			PC = 7;
			break;
        case 7:
			break;

        // vO`FW
        case 8: break;

        // Rg[`FW
        case 10: PC = 11; break;
        case 11:
			switch(MByte[0]) {
				case MIDI_CC_Moduration:
					MWBuf[Rch] = MByte[1] >> 2;
					break;
				case MIDI_CC_DataEntry:
					if((RPNMsbBuf[Rch] == 0) && (RPNLSBBuf[Rch] == 1))
						DetuneBuf[Rch] = MByte[1];
					break;
				case MIDI_CC_Control1:
					ShapeBuf[Rch] = MByte[1] >> 5;
					break;
				case MIDI_CC_Control2:
					RateBuf[Rch] = MByte[1] >> 4;
					break;
				case MIDI_CC_Control3:
					DepthBuf[Rch] = MByte[1] >> 5;
					break;
				case MIDI_CC_RPN_LSB:
					RPNLSBBuf[Rch] = MByte[1];
					break;
				case MIDI_CC_RPN_MSB:
					RPNMsbBuf[Rch] = MByte[1];
					break;
				case MIDI_MM_AllSoundOff:
					NoteCnt[Rch] = 0;
					Gate(OFF);
					break;
				case MIDI_MM_ResetAllControl:
					PBBuf[Rch] = 64;
					MWBuf[Rch] = 0;
					RateBuf[Rch] = 4;
					ShapeBuf[Rch] = 0;
					DepthBuf[Rch] = 0;
					RPNMsbBuf[Rch] = 127;
					RPNLSBBuf[Rch] = 127;
					DetuneBuf[Rch] = 64;
					break;
				case MIDI_MM_AllNoteOff:
					NoteCnt[Rch] = 0;
					Gate(OFF);
					break;
			}
			break;

        // `lvbV[
		case 12:
			PC = 13;
			break;
		case 13:
			break;

        // sb`xh
		case 14:
			PC = 15;
			break;
		case 15:
			PBBuf[Rch] = MByte[1];
			break;

		default:
			break;
    }
}

//-----------------------------------------------
// C

int16 main(void)
{
	uchar i;
	uchar c;

	DDRB = 0x3F;
	DDRC = 0x3F;
	DDRD = 0xFE;

	WR_DISABLE;
	LED_MIDI_OFF;
	LED_BEAT_OFF;
	SYNC_CLK_LOW;
	SYNC_STOP;

	Wait(300000);

	// CMU-800
	Init8255();
	Init8253();
	AllGateOff();

	// obt@ϐ
	for(i = 0; i < MAX_CH; i++)	{
		CVBuf[i] = 0x80;
		PBBuf[i] = 64;
		MWBuf[i] = 0;
		RateBuf[i] = 4;
		ShapeBuf[i] = 0;
		DepthBuf[i] = 0;
		NoteCnt[i] = 0;
		RPNMsbBuf[i] = 127;
		RPNLSBBuf[i] = 127;
		DetuneBuf[i] = 64;
	}

	// USART CLOCK=20MHz BAUD=31.25kHz
	// UBRR = (CLOCK / (16BAUD)) - 1 = 39
	UCSR0B = 0x00;
	UCSR0A = 0x00;
	UCSR0C = 0x06;	// 8bit STOP1bit
	UBRR0L = 0x27;	// 39
	UBRR0H = 0x00;	
	UCSR0B |= (1 << RXCIE0) | (1 << RXEN0);
	Ptr_buf_in = Ptr_buf_out = 0;

	// ^C}1ݒ
	TCCR1A = 0;
	TCCR1B = 3;
	TIFR1 |= (1 << TOV1);
	TIMSK1 |= (1<< TOIE1);
	TCNT1 = INTRCOUNT;

	// 荞݋
	sei();

	// C[v
	while(TRUE) {
		if(Ptr_buf_in == Ptr_buf_out) {
			if(Setflg) {
				Setflg = FALSE;
				// CVd`[W
				SetCV(c & 0x03);			// 1-4ch
				SetCV((c & 0x03) + 0x04);	// 5-8ch
				c++;
				for(i = 0; i < 6; i++) {
					SetDCO(i);
				}
			}
			continue;
		}

		// Oobt@PoCgo
		Ptr_buf_out++;
		Ptr_buf_out &= (BUFSIZE - 1);
		RxByte = RxBuf[Ptr_buf_out];

		if(MIDI_SystemMessage()) {
			MIDI_ChannelMessage();
		}
	}
	return(0);
}
